%% -*- erlang-indent-level: 4;indent-tabs-mode: nil; fill-column: 92 -*-
%% ex: ts=4 sw=4 et
%% @author Tim Dysinger <dysinger@chef.io>
%% Copyright 2012 Opscode, Inc. All Rights Reserved.
%%
%% This file is provided to you under the Apache License,
%% Version 2.0 (the "License"); you may not use this file
%% except in compliance with the License.  You may obtain
%% a copy of the License at
%%
%%   http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing,
%% software distributed under the License is distributed on an
%% "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
%% KIND, either express or implied.  See the License for the
%% specific language governing permissions and limitations
%% under the License.
%%

-module(bksw_xml).

-export([list_buckets/1, list_objects/2,
         signature_does_not_match_error/4,
         access_denied_error/1,
         model/0, write/1, write_erl/0, write_hrl/0]).

-include("amazon_s3.hrl").
-include("internal.hrl").

%%===================================================================
%% Public API
%%===================================================================

list_buckets(Buckets) ->
    #'ListAllMyBucketsResult'{'Owner' = owner(),
                              'Buckets' = buckets(Buckets)}.

list_objects(Bucket, none) ->
    list_objects(Bucket, []);
list_objects(Bucket, Objects) ->
    #'ListBucketResult'{'Name' = Bucket, 'Prefix' = "",
                        'Marker' = "", 'MaxKeys' = 10000, 'IsTruncated' = false,
                        'Contents' = [object(O1) || O1 <- Objects]}.

owner() ->
    #'CanonicalUser'{'ID' = "abc123",
                     'DisplayName' = "Bobo T. Clown"}.

buckets(Buckets) ->
    #'ListAllMyBucketsList'{'Bucket' =
                                [bucket(V1) || V1 <- Buckets]}.

bucket(#bucket{name = Name, date = Date}) ->
  #'ListAllMyBucketsEntry'{'Name' = Name,
                             'CreationDate' = bksw_format:to_date(Date)};
bucket(#db_bucket{bucket_name = Name, created_at = Date}) ->
    #'ListAllMyBucketsEntry'{'Name' = Name,
                             'CreationDate' = bksw_format:to_date(Date)}.

object(#db_file{name = Name, created_at = Date, data_size = Size,
                hash_md5 = Digest}) ->
     #'ListEntry'{'Key' = Name,
                 'LastModified' = bksw_format:to_date(Date),
                 'ETag' = bksw_format:to_etag(Digest),
                 'Size' = io_lib:format("~w", [Size]),
                 'Owner' = owner(),
                 'StorageClass' = "STANDARD"};
object(#object{name = Name, date = Date, size = Size,
               digest = Digest}) ->
    #'ListEntry'{'Key' = Name,
                 'LastModified' = bksw_format:to_date(Date),
                 'ETag' = bksw_format:to_etag(Digest),
                 'Size' = io_lib:format("~w", [Size]),
                 'Owner' = owner(),
                 'StorageClass' = "STANDARD"}.

signature_does_not_match_error(RequestId, SignatureProvided, StringToSign,
                               AWSAccessKeyId) ->
    Data =
        {'Error',
         [{'Code', ["SignatureDoesNotMatch"]},
          {'Message', ["The request signature we calculated does not match the "
                       "signature you provided. Check your key and signing method."]},
          {'StringToSignBytes', [""]},
          {'RequestId', [RequestId]},
      {'HostId', [""]},
      {'SignatureProvided', [SignatureProvided]},
      {'StringToSign', [StringToSign]},
      {'AWSAccessKeyId', [AWSAccessKeyId]}]},
    erlsom_ucs:to_utf8(xmerl:export_simple([Data], xmerl_xml,
                                           [{prolog,
                                             "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"}])).

access_denied_error(RequestId) ->
        Data =
        {'Error',
         [{'Code', ["AccessDenied"]},
          {'Message', ["Access Denied"]},
          {'RequestId', [RequestId]},
          {'HostId', [""]}]},
    erlsom_ucs:to_utf8(xmerl:export_simple([Data], xmerl_xml,
                                           [{prolog,
                                             "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"}])).
write(Xml) ->
    {ok, Text} = erlsom:write(Xml, model()),
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" ++
      erlsom_ucs:to_utf8(Text).

write_hrl() ->
    case
      file:read_file(bksw_util:file("amazon_s3.xsd"))
        of
      {ok, Bin} ->
          Hrl =
              erlsom_writeHrl:writeXsdHrlFile(erlsom_lib:toUnicode(Bin),
                                              []),
          file:write_file(bksw_util:file("../include/amazon_s3.hrl"),
                          Hrl);
      Error -> Error
    end.

write_erl() ->
    {ok, Model} =
        erlsom:compile_xsd_file(bksw_util:file("amazon_s3.xsd")),
    {ok, Device} =
        file:open(bksw_util:file("../src/bksw_xml.erl"),
                  [append, write]),
    ok = file:write(Device, io_lib:format("~p", [Model])),
    ok = file:close(Device).

model() ->
    {model,
     [{type, '_document', sequence,
       [{el,
         [{alt, 'CreateBucket', 'CreateBucket', [], 1, 1, true,
           undefined},
          {alt, 'CreateBucketResponse', 'CreateBucketResponse',
           [], 1, 1, true, undefined},
          {alt, 'DeleteBucket', 'DeleteBucket', [], 1, 1, true,
           undefined},
          {alt, 'DeleteBucketResponse', 'DeleteBucketResponse',
           [], 1, 1, true, undefined},
          {alt, 'GetBucketLoggingStatus',
           'GetBucketLoggingStatus', [], 1, 1, true, undefined},
          {alt, 'GetBucketLoggingStatusResponse',
           'GetBucketLoggingStatusResponse', [], 1, 1, true,
           undefined},
          {alt, 'SetBucketLoggingStatus',
           'SetBucketLoggingStatus', [], 1, 1, true, undefined},
          {alt, 'SetBucketLoggingStatusResponse',
           'SetBucketLoggingStatusResponse', [], 1, 1, true,
           undefined},
          {alt, 'GetObjectAccessControlPolicy',
           'GetObjectAccessControlPolicy', [], 1, 1, true,
           undefined},
          {alt, 'GetObjectAccessControlPolicyResponse',
           'GetObjectAccessControlPolicyResponse', [], 1, 1, true,
           undefined},
          {alt, 'GetBucketAccessControlPolicy',
           'GetBucketAccessControlPolicy', [], 1, 1, true,
           undefined},
          {alt, 'GetBucketAccessControlPolicyResponse',
           'GetBucketAccessControlPolicyResponse', [], 1, 1, true,
           undefined},
          {alt, 'SetObjectAccessControlPolicy',
           'SetObjectAccessControlPolicy', [], 1, 1, true,
           undefined},
          {alt, 'SetObjectAccessControlPolicyResponse',
           'SetObjectAccessControlPolicyResponse', [], 1, 1, true,
           undefined},
          {alt, 'SetBucketAccessControlPolicy',
           'SetBucketAccessControlPolicy', [], 1, 1, true,
           undefined},
          {alt, 'SetBucketAccessControlPolicyResponse',
           'SetBucketAccessControlPolicyResponse', [], 1, 1, true,
           undefined},
          {alt, 'GetObject', 'GetObject', [], 1, 1, true,
           undefined},
          {alt, 'GetObjectResponse', 'GetObjectResponse', [], 1,
           1, true, undefined},
          {alt, 'GetObjectExtended', 'GetObjectExtended', [], 1,
           1, true, undefined},
          {alt, 'GetObjectExtendedResponse',
           'GetObjectExtendedResponse', [], 1, 1, true, undefined},
          {alt, 'PutObject', 'PutObject', [], 1, 1, true,
           undefined},
          {alt, 'PutObjectResponse', 'PutObjectResponse', [], 1,
           1, true, undefined},
          {alt, 'PutObjectInline', 'PutObjectInline', [], 1, 1,
           true, undefined},
          {alt, 'PutObjectInlineResponse',
           'PutObjectInlineResponse', [], 1, 1, true, undefined},
          {alt, 'DeleteObject', 'DeleteObject', [], 1, 1, true,
           undefined},
          {alt, 'DeleteObjectResponse', 'DeleteObjectResponse',
           [], 1, 1, true, undefined},
          {alt, 'ListBucket', 'ListBucket', [], 1, 1, true,
           undefined},
          {alt, 'ListBucketResponse', 'ListBucketResponse', [], 1,
           1, true, undefined},
          {alt, 'ListVersionsResponse', 'ListVersionsResponse',
           [], 1, 1, true, undefined},
          {alt, 'ListAllMyBuckets', 'ListAllMyBuckets', [], 1, 1,
           true, undefined},
          {alt, 'ListAllMyBucketsResponse',
           'ListAllMyBucketsResponse', [], 1, 1, true, undefined},
          {alt, 'PostResponse', 'PostResponse', [], 1, 1, true,
           undefined},
          {alt, 'CopyObject', 'CopyObject', [], 1, 1, true,
           undefined},
          {alt, 'CopyObjectResponse', 'CopyObjectResponse', [], 1,
           1, true, undefined},
          {alt, 'ListAllMyBucketsResult',
           'ListAllMyBucketsResult', [], 1, 1, true, undefined},
          {alt, 'ListBucketResult', 'ListBucketResult', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 1, 1, 1, false, undefined},
      {type, 'TopicConfiguration', sequence,
       [{el,
         [{alt, 'Topic', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Event', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, unbound, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'NotificationConfiguration', sequence,
       [{el,
         [{alt, 'TopicConfiguration', 'TopicConfiguration', [],
           1, 1, true, undefined}],
         0, unbound, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'VersioningConfiguration', sequence,
       [{el,
         [{alt, 'Status', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 1},
        {el,
         [{alt, 'MfaDelete', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'RequestPaymentConfiguration', sequence,
       [{el,
         [{alt, 'Payer', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'CopyObjectResult', sequence,
       [{el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'CopyObjectResponse', sequence,
       [{el,
         [{alt, 'CopyObjectResult', 'CopyObjectResult', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'CopyObject', sequence,
       [{el,
         [{alt, 'SourceBucket', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'SourceKey', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'DestinationBucket', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'DestinationKey', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'MetadataDirective', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, 100, undefined, 6},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'CopySourceIfModifiedSince', {'#PCDATA', char},
           [], 1, 1, true, undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'CopySourceIfUnmodifiedSince', {'#PCDATA', char},
           [], 1, 1, true, undefined}],
         0, 1, undefined, 9},
        {el,
         [{alt, 'CopySourceIfMatch', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 100, undefined, 10},
        {el,
         [{alt, 'CopySourceIfNoneMatch', {'#PCDATA', char}, [],
           1, 1, true, undefined}],
         0, 100, undefined, 11},
        {el,
         [{alt, 'StorageClass', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 12},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 13},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 14},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 15},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 16}],
       [], undefined, undefined, 17, 1, 1, undefined,
       undefined},
      {type, 'PostResponse', sequence,
       [{el,
         [{alt, 'Location', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4}],
       [], undefined, undefined, 5, 1, 1, undefined,
       undefined},
      {type, 'ListAllMyBucketsList', sequence,
       [{el,
         [{alt, 'Bucket', 'ListAllMyBucketsEntry', [], 1, 1,
           true, undefined}],
         0, unbound, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'ListAllMyBucketsResult', sequence,
       [{el,
         [{alt, 'Owner', 'CanonicalUser', [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Buckets', 'ListAllMyBucketsList', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'ListAllMyBucketsEntry', sequence,
       [{el,
         [{alt, 'Name', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'CreationDate', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'ListAllMyBucketsResponse', sequence,
       [{el,
         [{alt, 'ListAllMyBucketsResponse',
           'ListAllMyBucketsResult', [], 1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'ListAllMyBuckets', sequence,
       [{el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 1},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3}],
       [], undefined, undefined, 4, 1, 1, undefined,
       undefined},
      {type, 'ListVersionsResult', sequence,
       [{el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 1},
        {el,
         [{alt, 'Name', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'Prefix', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'KeyMarker', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'VersionIdMarker', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 5},
        {el,
         [{alt, 'NextKeyMarker', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'NextVersionIdMarker', {'#PCDATA', char}, [], 1,
           1, true, undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'MaxKeys', {'#PCDATA', integer}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 8},
        {el,
         [{alt, 'Delimiter', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 9},
        {el,
         [{alt, 'IsTruncated', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 10},
        {el,
         [{alt, 'Version', 'VersionEntry', [], 1, 1, true,
           undefined},
          {alt, 'DeleteMarker', 'DeleteMarkerEntry', [], 1, 1,
           true, undefined}],
         0, unbound, undefined, 11},
        {el,
         [{alt, 'CommonPrefixes', 'PrefixEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 12}],
       [], undefined, undefined, 13, 1, 1, undefined,
       undefined},
      {type, 'ListBucketResult', sequence,
       [{el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 1},
        {el,
         [{alt, 'Name', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'Prefix', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'Marker', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'NextMarker', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'MaxKeys', {'#PCDATA', integer}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 6},
        {el,
         [{alt, 'Delimiter', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'IsTruncated', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 8},
        {el,
         [{alt, 'Contents', 'ListEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 9},
        {el,
         [{alt, 'CommonPrefixes', 'PrefixEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 10}],
       [], undefined, undefined, 11, 1, 1, undefined,
       undefined},
      {type, 'PrefixEntry', sequence,
       [{el,
         [{alt, 'Prefix', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'DeleteMarkerEntry', sequence,
       [{el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'VersionId', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'IsLatest', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'Owner', 'CanonicalUser', [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       undefined},
      {type, 'VersionEntry', sequence,
       [{el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'VersionId', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'IsLatest', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 5},
        {el,
         [{alt, 'Size', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 6},
        {el,
         [{alt, 'Owner', 'CanonicalUser', [], 1, 1, true,
           undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'StorageClass', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 8}],
       [], undefined, undefined, 9, 1, 1, undefined,
       undefined},
      {type, 'ListEntry', sequence,
       [{el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'Size', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'Owner', 'CanonicalUser', [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'StorageClass', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 6}],
       [], undefined, undefined, 7, 1, 1, undefined,
       undefined},
      {type, 'ListVersionsResponse', sequence,
       [{el,
         [{alt, 'ListVersionsResponse', 'ListVersionsResult', [],
           1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'ListBucketResponse', sequence,
       [{el,
         [{alt, 'ListBucketResponse', 'ListBucketResult', [], 1,
           1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'ListBucket', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Prefix', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Marker', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'MaxKeys', {'#PCDATA', integer}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Delimiter', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 9}],
       [], undefined, undefined, 10, 1, 1, undefined,
       undefined},
      {type, 'DeleteObjectResponse', sequence,
       [{el,
         [{alt, 'DeleteObjectResponse', 'Status', [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'DeleteObject', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 6}],
       [], undefined, undefined, 7, 1, 1, undefined,
       undefined},
      {type, 'PutObjectInlineResponse', sequence,
       [{el,
         [{alt, 'PutObjectInlineResponse', 'PutObjectResult', [],
           1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'PutObjectInline', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, 100, undefined, 3},
        {el,
         [{alt, 'Data', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'ContentLength', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 5},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'StorageClass', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 9},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 10},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 11}],
       [], undefined, undefined, 12, 1, 1, undefined,
       undefined},
      {type, 'PutObjectResult', sequence,
       [{el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'PutObjectResponse', sequence,
       [{el,
         [{alt, 'PutObjectResponse', 'PutObjectResult', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'PutObject', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, 100, undefined, 3},
        {el,
         [{alt, 'ContentLength', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'StorageClass', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 9},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 10}],
       [], undefined, undefined, 11, 1, 1, undefined,
       undefined},
      {type, 'GetObjectExtendedResponse', sequence,
       [{el,
         [{alt, 'GetObjectResponse', 'GetObjectResult', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'GetObjectExtended', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'GetMetadata', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'GetData', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'InlineData', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 5},
        {el,
         [{alt, 'ByteRangeStart', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'ByteRangeEnd', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'IfModifiedSince', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'IfUnmodifiedSince', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 9},
        {el,
         [{alt, 'IfMatch', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 100, undefined, 10},
        {el,
         [{alt, 'IfNoneMatch', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 100, undefined, 11},
        {el,
         [{alt, 'ReturnCompleteObjectOnConditionFailure',
           {'#PCDATA', bool}, [], 1, 1, true, undefined}],
         0, 1, undefined, 12},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 13},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 14},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 15},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 16}],
       [], undefined, undefined, 17, 1, 1, undefined,
       undefined},
      {type, 'GetObjectResult', sequence,
       [{el,
         [{alt, 'Status', 'Status', [], 1, 1, true, undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Metadata', 'MetadataEntry', [], 1, 1, true,
           undefined}],
         0, unbound, undefined, 2},
        {el,
         [{alt, 'Data', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, true, 3},
        {el,
         [{alt, 'LastModified', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'ETag', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       'GetObjectResult'},
      {type, 'GetObjectResponse', sequence,
       [{el,
         [{alt, 'GetObjectResponse', 'GetObjectResult', [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'GetObject', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'GetMetadata', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'GetData', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 4},
        {el,
         [{alt, 'InlineData', {'#PCDATA', bool}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 5},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 7},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 8},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 9}],
       [], undefined, undefined, 10, 1, 1, undefined,
       undefined},
      {type, 'SetBucketAccessControlPolicyResponse', sequence,
       [], [], undefined, undefined, 1, 1, 1, undefined,
       undefined},
      {type, 'SetBucketAccessControlPolicy', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 6}],
       [], undefined, undefined, 7, 1, 1, undefined,
       undefined},
      {type, 'SetObjectAccessControlPolicyResponse', sequence,
       [], [], undefined, undefined, 1, 1, 1, undefined,
       undefined},
      {type, 'SetObjectAccessControlPolicy', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         1, 1, undefined, 3},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 6},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 7}],
       [], undefined, undefined, 8, 1, 1, undefined,
       undefined},
      {type, 'AccessControlPolicy', sequence,
       [{el,
         [{alt, 'Owner', 'CanonicalUser', [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'LocationConstraint', sequence,
       [{el,
         [{alt, '#text', {'#PCDATA', char}, [], 1, 1, false,
           undefined}],
         0, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'CreateBucketConfiguration', sequence,
       [{el,
         [{alt, 'LocationConstraint', 'LocationConstraint', [],
           1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'AccessControlList', sequence,
       [{el,
         [{alt, 'Grant', 'Grant', [], 1, 1, true, undefined}], 0,
         100, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'Grant', sequence,
       [{el,
         [{alt, 'Grantee', 'Grantee', [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Permission', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'Group', sequence,
       [{el,
         [{alt, 'URI', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined, 'Group'},
      {type, 'CanonicalUser', sequence,
       [{el,
         [{alt, 'ID', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'DisplayName', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       'CanonicalUser'},
      {type, 'AmazonCustomerByEmail', sequence,
       [{el,
         [{alt, 'EmailAddress', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       'AmazonCustomerByEmail'},
      {type, 'User', sequence, [], [], undefined, undefined,
       1, 1, 1, undefined, 'User'},
      {type, 'Grantee', sequence, [], [], undefined,
       undefined, 1, 1, 1, undefined, undefined},
      {type, 'GetBucketAccessControlPolicyResponse', sequence,
       [{el,
         [{alt, 'GetBucketAccessControlPolicyResponse',
           'AccessControlPolicy', [], 1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'GetBucketAccessControlPolicy', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       undefined},
      {type, 'GetObjectAccessControlPolicyResponse', sequence,
       [{el,
         [{alt, 'GetObjectAccessControlPolicyResponse',
           'AccessControlPolicy', [], 1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'GetObjectAccessControlPolicy', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Key', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 6}],
       [], undefined, undefined, 7, 1, 1, undefined,
       undefined},
      {type, 'SetBucketLoggingStatusResponse', sequence, [],
       [], undefined, undefined, 1, 1, 1, undefined,
       undefined},
      {type, 'SetBucketLoggingStatus', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5},
        {el,
         [{alt, 'BucketLoggingStatus', 'BucketLoggingStatus', [],
           1, 1, true, undefined}],
         1, 1, undefined, 6}],
       [], undefined, undefined, 7, 1, 1, undefined,
       undefined},
      {type, 'GetBucketLoggingStatusResponse', sequence,
       [{el,
         [{alt, 'GetBucketLoggingStatusResponse',
           'BucketLoggingStatus', [], 1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'GetBucketLoggingStatus', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       undefined},
      {type, 'LoggingSettings', sequence,
       [{el,
         [{alt, 'TargetBucket', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'TargetPrefix', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         1, 1, undefined, 2},
        {el,
         [{alt, 'TargetGrants', 'AccessControlList', [], 1, 1,
           true, undefined}],
         0, 1, undefined, 3}],
       [], undefined, undefined, 4, 1, 1, undefined,
       undefined},
      {type, 'BucketLoggingStatus', sequence,
       [{el,
         [{alt, 'LoggingEnabled', 'LoggingSettings', [], 1, 1,
           true, undefined}],
         0, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'DeleteBucketResponse', sequence,
       [{el,
         [{alt, 'DeleteBucketResponse', 'Status', [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'DeleteBucket', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Credential', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       undefined},
      {type, 'CreateBucketResult', sequence,
       [{el,
         [{alt, 'BucketName', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'Result', sequence,
       [{el,
         [{alt, 'Status', 'Status', [], 1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'Status', sequence,
       [{el,
         [{alt, 'Code', {'#PCDATA', integer}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Description', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'CreateBucketResponse', sequence,
       [{el,
         [{alt, 'CreateBucketReturn', 'CreateBucketResult', [],
           1, 1, true, undefined}],
         1, 1, undefined, 1}],
       [], undefined, undefined, 2, 1, 1, undefined,
       undefined},
      {type, 'MetadataEntry', sequence,
       [{el,
         [{alt, 'Name', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'Value', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 2}],
       [], undefined, undefined, 3, 1, 1, undefined,
       undefined},
      {type, 'CreateBucket', sequence,
       [{el,
         [{alt, 'Bucket', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         1, 1, undefined, 1},
        {el,
         [{alt, 'AccessControlList', 'AccessControlList', [], 1,
           1, true, undefined}],
         0, 1, undefined, 2},
        {el,
         [{alt, 'AWSAccessKeyId', {'#PCDATA', char}, [], 1, 1,
           true, undefined}],
         0, 1, undefined, 3},
        {el,
         [{alt, 'Timestamp', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 4},
        {el,
         [{alt, 'Signature', {'#PCDATA', char}, [], 1, 1, true,
           undefined}],
         0, 1, undefined, 5}],
       [], undefined, undefined, 6, 1, 1, undefined,
       undefined}],
     [{ns, "http://s3.amazonaws.com/doc/2006-03-01/",
       undefined},
      {ns, "http://www.w3.org/2001/XMLSchema", "xsd"}],
     "http://s3.amazonaws.com/doc/2006-03-01/",
     [{'GetObjectResult', 'Result'}, {'Group', 'Grantee'},
      {'CanonicalUser', 'User'},
      {'AmazonCustomerByEmail', 'User'},
      {'User', 'Grantee'}]}.
